import base64
from typing import Any, Dict


def urlsafeEncode(data: bytes) -> str:
    """
    Encode the given bytes to a URL-safe string.
    """
    return base64.urlsafe_b64encode(data).decode("ascii").rstrip("=")


def removeFalsyKeys(data: Dict[str, Any]) -> Dict[str, Any]:
    """
    Remove keys with falsy values from the given dictionary.
    """
    return {k: v for k, v in data.items() if v}



from typing import Any, Optional, Dict
from pydantic import BaseModel


class TokenResponse(BaseModel):
    """
    The response model from the token endpoint.
    """

    access_token: str
    """
    The access token string.
    """
    token_type: str
    """
    The token type string, should be "Bearer".
    """
    expires_in: int
    """
    The expiration time of the access token (in seconds).
    """
    refresh_token: Optional[str] = None
    """
    The refresh token string.
    """
    id_token: Optional[str] = None
    """
    The ID token string.
    """


class UserIdentity(BaseModel):
    """
    The user identity model.
    """

    userId: str
    """
    The user ID of the target identity.
    """
    details: Optional[Dict[str, Any]] = None
    """
    The details of the target identity, can be any JSON object.
    """


class UserInfoResponse(BaseModel):
    """
    The response model from the user info endpoint.
    """

    sub: str
    """
    The subject identifier for whom the token is intended (user ID).
    """
    name: Optional[str] = None
    """
    The full name of the user.
    """
    username: Optional[str] = None
    """
    The username of the user.
    """
    picture: Optional[str] = None
    """
    The profile picture URL of the user.
    """
    email: Optional[str] = None
    """
    The email address of the user.
    """
    email_verified: Optional[bool] = None
    """
    Whether the email address is verified.
    """
    phone_number: Optional[str] = None
    """
    The phone number of the user.
    """
    phone_number_verified: Optional[bool] = None
    """
    Whether the phone number is verified.
    """
    custom_data: Any = None
    """
    The custom data of the user, can be any JSON object.
    """
    identities: Optional[Dict[str, UserIdentity]] = None
    """
    The identities of the user, can be a dictionary of key-value pairs, where the key is
    the identity type and the value is the `UserIdentity` object.
    """



"""
The Logto client class and the related models.
"""

import time
from typing import Dict, List, Literal, Optional, Union
from pydantic import BaseModel
import urllib.parse

from logto.models.oidc import Scope

from .Storage import MemoryStorage, Storage
from .LogtoException import LogtoException
from .OidcCore import (
    AccessTokenClaims,
    IdTokenClaims,
    OidcCore,
    TokenResponse,
    UserInfoResponse,
)
from .utilities import removeFalsyKeys


class LogtoConfig(BaseModel):
    """
    The configuration object for the Logto client.
    """

    endpoint: str
    """
    The endpoint for the Logto server, you can get it from the integration guide
    or the team settings page of the Logto Console.

    Example:
    https://foo.logto.app
    """

    appId: str
    """
    The client ID of your application, you can get it from the integration guide
    or the application details page of the Logto Console.
    """

    appSecret: Optional[str] = None
    """
    The client secret of your application, you can get it from the integration guide
    or the application details page of the Logto Console.
    """

    prompt: Literal["consent", "login"] = "consent"
    """
    The prompt parameter for the OpenID Connect authorization request.

    - If the value is `consent`, the user will be able to reuse the existing consent
    without being prompted for sign-in again.
    - If the value is `login`, the user will be prompted for sign-in again anyway. Note
    there will be no Refresh Token returned in this case.
    """

    resources: List[str] = []
    """
    The API resources that your application needs to access. You can specify multiple
    resources by providing an array of strings.

    See https://docs.logto.io/docs/recipes/rbac/ to learn more about how to use role-based
    access control (RBAC) to protect API resources.
    """

    scopes: List[Union[str, Scope]] = []
    """
    The scopes (permissions) that your application needs to access.
    Scopes that will be added by default: `openid`, `offline_access` and `profile`.

    If resources are specified, scopes will be applied to every resource.

    See https://docs.logto.io/docs/recipes/integrate-logto/vanilla-js/#fetch-user-information
    for more information of available scopes for user information.
    """


class SignInSession(BaseModel):
    """
    The sign-in session that stores the information for the sign-in callback.
    Should be stored before redirecting the user to Logto.
    """

    redirectUri: str
    """
    The redirect URI for the current sign-in session.
    """
    codeVerifier: str
    """
    The code verifier of Proof Key for Code Exchange (PKCE).
    """
    state: str
    """
    The state for OAuth 2.0 authorization request.
    """


class AccessToken(BaseModel):
    """
    The access token class for a resource.
    """

    token: str
    """
    The access token string.
    """
    expiresAt: int
    """
    The timestamp (in seconds) when the access token will expire.
    Note this is not the expiration time of the access token itself, but the
    expiration time of the access token cache.
    """


class AccessTokenMap(BaseModel):
    """
    The access token map that maps the resource to the access token for that resource.

    If resource is an empty string, it means the access token is for UserInfo endpoint
    or the default resource.
    """

    x: Dict[str, AccessToken]


InteractionMode = Literal["signIn", "signUp"]
"""
The interaction mode for the sign-in request. Note this is not a part of the OIDC
specification, but a Logto extension.
"""


class LogtoClient:
    """
    The main class of the Logto client. You should create an instance of this class
    and use it to sign in, sign out, get access token, etc.
    """

    def __init__(self, config: LogtoConfig, storage: Storage = MemoryStorage()) -> None:
        self.config = config
        self._oidcCore: Optional[OidcCore] = None
        self._storage = storage

    async def getOidcCore(self) -> OidcCore:
        """
        Get the OIDC core object. You can use it to get the provider metadata, verify
        the ID token, fetch tokens by code or refresh token, etc.
        """
        if self._oidcCore is None:
            self._oidcCore = OidcCore(
                await OidcCore.getProviderMetadata(
                    f"{self.config.endpoint}/oidc/.well-known/openid-configuration"
                )
            )
        return self._oidcCore

    def _getAccessTokenMap(self) -> AccessTokenMap:
        """
        Get the access token map from storage.
        """
        accessTokenMap = self._storage.get("accessTokenMap")
        try:
            return AccessTokenMap.model_validate_json(accessTokenMap)
        except:
            return AccessTokenMap(x={})

    def _setAccessToken(self, resource: str, accessToken: str, expiresIn: int) -> None:
        """
        Set the access token for the given resource to storage.
        """
        accessTokenMap = self._getAccessTokenMap()
        accessTokenMap.x[resource] = AccessToken(
            token=accessToken,
            expiresAt=int(time.time())
            + expiresIn
            - 60,  # 60 seconds earlier to avoid clock skew
        )
        self._storage.set("accessTokenMap", accessTokenMap.model_dump_json())

    def _getAccessToken(self, resource: str) -> Optional[str]:
        """
        Get the valid access token for the given resource from storage, no refresh will be
        performed.
        """
        accessTokenMap = self._getAccessTokenMap()
        accessToken = accessTokenMap.x.get(resource, None)
        if accessToken is None or accessToken.expiresAt < int(time.time()):
            return None
        return accessToken.token

    async def _handleTokenResponse(
        self, resource: str, tokenResponse: TokenResponse
    ) -> None:
        """
        Handle the token response from the Logto server and store the tokens to storage.

        Resource can be an empty string, which means the access token is for UserInfo
        endpoint or the default resource.
        """
        if tokenResponse.id_token is not None:
            (await self.getOidcCore()).verifyIdToken(
                tokenResponse.id_token, self.config.appId
            )
            self._storage.set("idToken", tokenResponse.id_token)

        if tokenResponse.refresh_token is not None:
            self._storage.set("refreshToken", tokenResponse.refresh_token)

        self._setAccessToken(
            resource, tokenResponse.access_token, tokenResponse.expires_in
        )

    async def _buildSignInUrl(
        self,
        redirectUri: str,
        codeChallenge: str,
        state: str,
        interactionMode: Optional[InteractionMode] = None,
    ) -> str:
        appId, prompt, resources, scopes = (
            self.config.appId,
            self.config.prompt,
            self.config.resources,
            self.config.scopes,
        )
        authorizationEndpoint = (
            await self.getOidcCore()
        ).metadata.authorization_endpoint
        query = urllib.parse.urlencode(
            removeFalsyKeys(
                {
                    "client_id": appId,
                    "redirect_uri": redirectUri,
                    "response_type": "code",
                    "scope": " ".join(
                        (item.value if isinstance(item, Scope) else item)
                        for item in (scopes + OidcCore.defaultScopes)
                    ),
                    "resource": resources,
                    "prompt": prompt,
                    "code_challenge": codeChallenge,
                    "code_challenge_method": "S256",
                    "state": state,
                    "interaction_mode": interactionMode,
                }
            ),
            True,
        )
        return f"{authorizationEndpoint}?{query}"

    def _getSignInSession(self) -> Optional[SignInSession]:
        """
        Try to parse the current sign-in session from storage. If the value does not
        exist or parse failed, return None.
        """
        signInSession = self._storage.get("signInSession")
        if signInSession is None:
            return None
        try:
            return SignInSession.model_validate_json(signInSession)
        except:
            return None

    def _setSignInSession(self, signInSession: SignInSession) -> None:
        self._storage.set("signInSession", signInSession.model_dump_json())

    async def signIn(
        self, redirectUri: str, interactionMode: Optional[InteractionMode] = None
    ) -> str:
        """
        Returns the sign-in URL for the given redirect URI. You should redirect the user
        to the returned URL to sign in.

        By specifying the interaction mode, you can control whether the user will be
        prompted for sign-in or sign-up on the first screen. If the interaction mode is
        not specified, the default one will be used.

        Example:
          ```python
          return redirect(await client.signIn('https://example.com/callback'))
          ```
        """
        codeVerifier = OidcCore.generateCodeVerifier()
        codeChallenge = OidcCore.generateCodeChallenge(codeVerifier)
        state = OidcCore.generateState()
        signInUrl = await self._buildSignInUrl(
            redirectUri, codeChallenge, state, interactionMode
        )

        self._setSignInSession(
            SignInSession(
                redirectUri=redirectUri,
                codeVerifier=codeVerifier,
                state=state,
            )
        )
        for key in ["idToken", "accessToken", "refreshToken"]:
            self._storage.delete(key)

        return signInUrl

    async def signOut(self, postLogoutRedirectUri: Optional[str] = None) -> str:
        """
        Returns the sign-out URL for the given post-logout redirect URI. You should
        redirect the user to the returned URL to sign out.

        If the post-logout redirect URI is not provided, the Logto default post-logout
        redirect URI will be used.

        Note:
          If the OpenID Connect server does not support the end session endpoint
          (i.e. OpenID Connect RP-Initiated Logout), the function will throw an
          exception. Logto supports the end session endpoint.

        Example:
          ```python
          return redirect(await client.signOut('https://example.com'))
          ```
        """
        self._storage.delete("idToken")
        self._storage.delete("refreshToken")
        self._storage.delete("accessTokenMap")

        endSessionEndpoint = (await self.getOidcCore()).metadata.end_session_endpoint

        if endSessionEndpoint is None:
            raise LogtoException(
                "End session endpoint not found in the provider metadata"
            )

        return (
            endSessionEndpoint
            + "?"
            + urllib.parse.urlencode(
                removeFalsyKeys(
                    {
                        "client_id": self.config.appId,
                        "post_logout_redirect_uri": postLogoutRedirectUri,
                    }
                )
            )
        )

    async def handleSignInCallback(self, callbackUri: str) -> None:
        """
        Handle the sign-in callback from the Logto server. This method should be called
        in the callback route handler of your application.
        """
        signInSession = self._getSignInSession()

        if signInSession is None:
            raise LogtoException("Sign-in session not found")

        # Validate the callback URI without query matches the redirect URI
        parsedCallbackUri = urllib.parse.urlparse(callbackUri)

        if (
            parsedCallbackUri.path
            != urllib.parse.urlparse(signInSession.redirectUri).path
        ):
            raise LogtoException(
                "The URI path does not match the redirect URI in the sign-in session"
            )

        query = urllib.parse.parse_qs(parsedCallbackUri.query)

        if "error" in query:
            raise LogtoException(query["error"][0])

        if signInSession.state != query.get("state", [None])[0]:
            raise LogtoException("Invalid state in the callback URI")

        code = query.get("code", [None])[0]
        if code is None:
            raise LogtoException("Code not found in the callback URI")

        tokenResponse = await (await self.getOidcCore()).fetchTokenByCode(
            clientId=self.config.appId,
            clientSecret=self.config.appSecret,
            redirectUri=signInSession.redirectUri,
            code=code,
            codeVerifier=signInSession.codeVerifier,
        )

        await self._handleTokenResponse("", tokenResponse)
        self._storage.delete("signInSession")

    async def getAccessToken(self, resource: str = "") -> Optional[str]:
        """
        Get the access token for the given resource. If the access token is expired,
        it will be refreshed automatically. If no refresh token is found, None will
        be returned.
        """
        accessToken = self._getAccessToken(resource)
        if accessToken is not None:
            return accessToken

        refreshToken = self._storage.get("refreshToken")
        if refreshToken is None:
            return None

        tokenResponse = await (await self.getOidcCore()).fetchTokenByRefreshToken(
            clientId=self.config.appId,
            clientSecret=self.config.appSecret,
            refreshToken=refreshToken,
            resource=resource,
        )

        await self._handleTokenResponse(resource, tokenResponse)
        return tokenResponse.access_token

    async def getAccessTokenClaims(self, resource: str = "") -> AccessTokenClaims:
        """
        Get the claims in the access token for the given resource. If the access token
        is expired, it will be refreshed automatically. If it's unable to refresh the
        access token, an exception will be thrown.
        """
        accessToken = await self.getAccessToken(resource)
        return OidcCore.decodeAccessToken(accessToken)

    def getIdToken(self) -> Optional[str]:
        """
        Get the ID Token string. If you need to get the claims in the ID Token, use
        `getIdTokenClaims` instead.
        """
        return self._storage.get("idToken")

    def getIdTokenClaims(self) -> IdTokenClaims:
        """
        Get the claims in the ID Token. If the ID Token does not exist, an exception
        will be thrown.
        """
        idToken = self._storage.get("idToken")
        if idToken is None:
            raise LogtoException("ID Token not found")

        return OidcCore.decodeIdToken(idToken)

    def getRefreshToken(self) -> Optional[str]:
        """
        Get the refresh token string.
        """
        return self._storage.get("refreshToken")

    def isAuthenticated(self) -> bool:
        """
        Check if the user is authenticated by checking if the ID Token exists.
        """
        return self._storage.get("idToken") is not None

    async def fetchUserInfo(self) -> UserInfoResponse:
        """
        Fetch the user information from the UserInfo endpoint. If the access token
        is expired, it will be refreshed automatically.
        """
        accessToken = await self.getAccessToken()
        return await (await self.getOidcCore()).fetchUserInfo(accessToken)



from typing import Any, Callable, Dict, Optional
from pytest_mock import MockerFixture
import pytest

from . import LogtoClient, LogtoConfig, LogtoException, Storage
from .utilities.test import mockHttp, mockProviderMetadata
from .models.response import TokenResponse, UserInfoResponse
from .models.oidc import IdTokenClaims, AccessTokenClaims, UserInfoScope
from .Storage import MemoryStorage, Storage
from .OidcCore import OidcCore

MockRequest = Callable[..., None]


class TestLogtoClient:
    @pytest.fixture
    def config(self) -> LogtoConfig:
        return LogtoConfig(
            endpoint="http://localhost:3001",  # Replace with your Logto endpoint
            appId="replace-with-your-app-id",
        )

    @pytest.fixture
    def mockRequest(self, mocker: MockerFixture) -> MockRequest:
        def _mock(
            method: str = "get",
            json: Optional[Dict[str, Any]] = None,
            text: Optional[str] = None,
            status: int = 200,
        ):
            return mockHttp(mocker, method, json, text, status)

        return _mock

    @pytest.fixture
    def storage(self) -> Storage:
        return MemoryStorage()

    @pytest.fixture
    def client(
        self,
        config: LogtoConfig,
        storage: Storage,
        mockRequest: MockRequest,
        mocker: MockerFixture,
    ) -> LogtoClient:
        mocker.patch(
            "logto.OidcCore.OidcCore.generateCodeVerifier", return_value="codeVerifier"
        )
        mocker.patch(
            "logto.OidcCore.OidcCore.generateCodeChallenge",
            return_value="codeChallenge",
        )
        mocker.patch("logto.OidcCore.OidcCore.generateState", return_value="state")
        mockRequest(json=mockProviderMetadata.__dict__)
        return LogtoClient(config, storage)

    async def test_getOidcCore(self, client: LogtoClient) -> None:
        assert isinstance(await client.getOidcCore(), OidcCore)

    async def test_signIn(self, client: LogtoClient) -> None:
        url = await client.signIn("redirectUri", "signUp")

        assert (
            url
            == "https://logto.app/oidc/auth?client_id=replace-with-your-app-id&redirect_uri=redirectUri&response_type=code&scope=openid+offline_access+profile&prompt=consent&code_challenge=codeChallenge&code_challenge_method=S256&state=state&interaction_mode=signUp"
        )

    async def test_signIn_multipleResources(self, client: LogtoClient) -> None:
        client.config.resources = ["https://resource1", "https://resource2"]
        url = await client.signIn("redirectUri", "signUp")

        assert (
            url
            == "https://logto.app/oidc/auth?client_id=replace-with-your-app-id&redirect_uri=redirectUri&response_type=code&scope=openid+offline_access+profile&resource=https%3A%2F%2Fresource1&resource=https%3A%2F%2Fresource2&prompt=consent&code_challenge=codeChallenge&code_challenge_method=S256&state=state&interaction_mode=signUp"
        )

    async def test_signIn_multipleScopes(self, client: LogtoClient) -> None:
        client.config.scopes = [UserInfoScope.email, "phone"]
        url = await client.signIn("redirectUri")

        assert (
            url
            == "https://logto.app/oidc/auth?client_id=replace-with-your-app-id&redirect_uri=redirectUri&response_type=code&scope=email+phone+openid+offline_access+profile&prompt=consent&code_challenge=codeChallenge&code_challenge_method=S256&state=state"
        )

    async def test_signIn_allConfigs(self, client: LogtoClient) -> None:
        client.config.scopes = ["email", "phone"]
        client.config.resources = ["https://resource1", "https://resource2"]
        client.config.prompt = "login"
        url = await client.signIn("redirectUri", "signUp")

        assert (
            url
            == "https://logto.app/oidc/auth?client_id=replace-with-your-app-id&redirect_uri=redirectUri&response_type=code&scope=email+phone+openid+offline_access+profile&resource=https%3A%2F%2Fresource1&resource=https%3A%2F%2Fresource2&prompt=login&code_challenge=codeChallenge&code_challenge_method=S256&state=state&interaction_mode=signUp"
        )

    async def test_signOut(
        self, client: LogtoClient, storage: Storage, mockRequest: MockRequest
    ) -> None:
        # Add end session endpoint to metadata
        mockRequest(
            method="get",
            json={
                **mockProviderMetadata.__dict__,
                "end_session_endpoint": "https://logto.app/oidc/logout",
            },
        )

        storage.set("idToken", "idToken")
        storage.set("accessTokenMap", "accessTokenMap")
        storage.set("refreshToken", "refreshToken")

        url = await client.signOut("redirectUri")

        assert (
            url
            == "https://logto.app/oidc/logout?client_id=replace-with-your-app-id&post_logout_redirect_uri=redirectUri"
        )

        assert storage.get("idToken") is None
        assert storage.get("accessTokenMap") is None
        assert storage.get("refreshToken") is None

    async def test_signOut_failure(self, client: LogtoClient) -> None:
        with pytest.raises(
            LogtoException,
            match="End session endpoint not found in the provider metadata",
        ):
            await client.signOut("redirectUri")

    async def test_handleSignInCallback_sessionNotFound(
        self,
        client: LogtoClient,
    ) -> None:
        with pytest.raises(LogtoException, match="Sign-in session not found"):
            await client.handleSignInCallback(callbackUri="https://redirect_uri")

    async def test_handleSignInCallback_pathDoesNotMatch(
        self, client: LogtoClient, storage: Storage
    ) -> None:
        storage.set(
            "signInSession",
            '{"redirectUri": "https://redirect_uri/some_path", "codeVerifier": "codeVerifier", "state": "state"}',
        )
        with pytest.raises(
            LogtoException,
            match="The URI path does not match the redirect URI in the sign-in session",
        ):
            await client.handleSignInCallback(callbackUri="https://redirect_uri")

    async def test_handleSignInCallback_stateDoesNotMatch(
        self, client: LogtoClient, storage: Storage
    ) -> None:
        storage.set(
            "signInSession",
            '{"redirectUri": "https://redirect_uri", "codeVerifier": "codeVerifier", "state": "state"}',
        )
        with pytest.raises(
            LogtoException,
            match="Invalid state in the callback URI",
        ):
            await client.handleSignInCallback(
                callbackUri="https://redirect_uri?state=state2"
            )

    async def test_handleSignInCallback_codeNotFound(
        self, client: LogtoClient, storage: Storage
    ) -> None:
        storage.set(
            "signInSession",
            '{"redirectUri": "https://redirect_uri", "codeVerifier": "codeVerifier", "state": "state"}',
        )
        with pytest.raises(LogtoException, match="Code not found in the callback URI"):
            await client.handleSignInCallback(
                callbackUri="https://redirect_uri?state=state"
            )

    async def test_handleSignInCallback(
        self,
        client: LogtoClient,
        storage: Storage,
        mockRequest: MockRequest,
        mocker: MockerFixture,
    ) -> None:
        storage.set(
            "signInSession",
            '{"redirectUri": "https://redirect_uri", "codeVerifier": "codeVerifier", "state": "state"}',
        )

        # Mock getOidcCore()
        client.getOidcCore = mocker.AsyncMock(
            return_value=OidcCore(mockProviderMetadata),
        )

        # Mock token response
        tokenResponse = TokenResponse(
            access_token="accessToken",
            token_type="Bearer",
            expires_in=3600,
            refresh_token="refreshToken",
            id_token="idToken",
        )

        mockRequest(method="post", json=tokenResponse.__dict__)

        # Mock verifyIdToken()
        mocker.patch("logto.OidcCore.OidcCore.verifyIdToken", return_value=None)

        # Should not raise
        await client.handleSignInCallback(
            callbackUri="https://redirect_uri?state=state&code=code"
        )

        assert storage.get("idToken") == "idToken"
        assert storage.get("refreshToken") == "refreshToken"
        assert await client.getAccessToken() == "accessToken"

    async def test_getAccessToken_cached(
        self,
        client: LogtoClient,
        storage: Storage,
    ) -> None:
        assert await client.getAccessToken() == None
        storage.set(
            "accessTokenMap",
            '{"x":{"":{"token":"access_token","expiresAt": 9999999999}, "foo":{"token":"access_token_foo","expiresAt": 9999999999}}}',
        )
        assert await client.getAccessToken() == "access_token"
        assert await client.getAccessToken(resource="foo") == "access_token_foo"

    async def test_getAccessToken_noRefreshToken(
        self,
        client: LogtoClient,
        storage: Storage,
    ) -> None:
        storage.set("accessTokenMap", '{"x":{}}')
        storage.set("refreshToken", None)
        assert await client.getAccessToken() == None

    async def test_getAccessToken_useRefreshToken(
        self,
        client: LogtoClient,
        storage: Storage,
        mockRequest: MockRequest,
    ) -> None:
        storage.set("accessTokenMap", '{"x":{}}')
        storage.set("refreshToken", "refreshToken")

        # Mock token response
        tokenResponse = TokenResponse(
            access_token="accessToken",
            token_type="Bearer",
            expires_in=3600,
            refresh_token="refreshToken",
        )
        mockRequest(method="post", json=tokenResponse.__dict__)

        assert await client.getAccessToken() == "accessToken"

    async def test_getAccessTokenClaims(
        self, client: LogtoClient, storage: Storage
    ) -> None:
        # Assign a valid access token raw string
        accessToken = "eyJhbGciOiJSUzI1NiIsImtpZCI6IjEifQ.eyJpc3MiOiJodHRwczovL2xvZ3RvLmFwcCIsImF1ZCI6Imh0dHBzOi8vbG9ndG8uYXBwL2FwaSIsImV4cCI6OTk5OTk5OTk5OSwiaWF0IjoxNjE2NDQ2MzAwLCJzdWIiOiJ1c2VyMSIsInNjb3BlIjoiYWRtaW4gdXNlciIsImNsaWVudF9pZCI6InNhcXJlMW9xYmtwajZ6aHE4NWhvMCJ9.12345678901234567890123456789012345678901234567890"
        storage.set(
            "accessTokenMap",
            '{"x":{"":{"token":"' + accessToken + '", "expiresAt": 9999999999}}}',
        )

        assert await client.getAccessTokenClaims() == AccessTokenClaims(
            iss="https://logto.app",
            aud="https://logto.app/api",
            exp=9999999999,
            iat=1616446300,
            sub="user1",
            scope="admin user",
            client_id="saqre1oqbkpj6zhq85ho0",
        )

    async def test_getIdToken(
        self,
        client: LogtoClient,
        storage: Storage,
    ) -> None:
        assert client.getIdToken() == None
        storage.set("idToken", "idToken")
        assert client.getIdToken() == "idToken"

    async def test_getIdTokenClaims(
        self, client: LogtoClient, storage: Storage
    ) -> None:
        idTokenString = "eyJhbGciOiJSUzI1NiIsImtpZCI6IjEifQ.eyJpc3MiOiJodHRwczovL2xvZ3RvLmFwcCIsImF1ZCI6ImZvbyIsImV4cCI6MTYxNjQ0NjQwMCwiaWF0IjoxNjE2NDQ2MzAwLCJzdWIiOiJ1c2VyMSIsIm5hbWUiOiJKb2huIFdpY2siLCJ1c2VybmFtZSI6ImpvaG4iLCJlbWFpbCI6ImpvaG5Ad2ljay5jb20iLCJlbWFpbF92ZXJpZmllZCI6dHJ1ZX0.12345678901234567890123456789012345678901234567890"
        storage.set("idToken", idTokenString)
        assert client.getIdTokenClaims() == IdTokenClaims(
            iss="https://logto.app",
            aud="foo",
            exp=1616446400,
            iat=1616446300,
            sub="user1",
            name="John Wick",
            username="john",
            email="john@wick.com",
            email_verified=True,
        )

    async def test_getRefreshToken(
        self,
        client: LogtoClient,
        storage: Storage,
    ) -> None:
        assert client.getRefreshToken() == None
        storage.set("refreshToken", "refreshToken")
        assert client.getRefreshToken() == "refreshToken"

    async def test_isAuthenticated(
        self,
        client: LogtoClient,
        storage: Storage,
    ) -> None:
        assert client.isAuthenticated() == False
        storage.set("idToken", "idToken")
        assert client.isAuthenticated() == True

    async def test_fetchUserInfo(
        self, client: LogtoClient, mocker: MockerFixture
    ) -> None:
        userinfoResponse = UserInfoResponse(
            sub="user1", name="John Wick", username="john", email="john@wick.com"
        )

        async def mockFetchUserInfo(accessToken: str) -> UserInfoResponse:
            return userinfoResponse

        client.getAccessToken = mocker.AsyncMock(return_value="accessToken")
        mocker.patch(
            "logto.OidcCore.OidcCore.fetchUserInfo",
            side_effect=mockFetchUserInfo,
        )

        assert await client.fetchUserInfo() == userinfoResponse



from enum import Enum
from typing import List, Optional
from pydantic import BaseModel, ConfigDict


class OidcProviderMetadata(BaseModel):
    """
    The OpenID Connect Discovery response object.

    See https://openid.net/specs/openid-connect-discovery-1_0.html#ProviderMetadata
    """

    issuer: str
    authorization_endpoint: str
    token_endpoint: str
    userinfo_endpoint: str  # This is actually "RECOMMENDED" but Logto requires it
    jwks_uri: str
    registration_endpoint: Optional[str] = None
    scopes_supported: List[str] = []
    response_types_supported: List[str]
    response_modes_supported: List[str] = []
    grant_types_supported: List[str] = []
    acr_values_supported: List[str] = []
    subject_types_supported: List[str]
    id_token_signing_alg_values_supported: List[str]
    id_token_encryption_alg_values_supported: List[str] = []
    id_token_encryption_enc_values_supported: List[str] = []
    userinfo_signing_alg_values_supported: List[str] = []
    userinfo_encryption_alg_values_supported: List[str] = []
    userinfo_encryption_enc_values_supported: List[str] = []
    request_object_signing_alg_values_supported: List[str] = []
    request_object_encryption_alg_values_supported: List[str] = []
    request_object_encryption_enc_values_supported: List[str] = []
    token_endpoint_auth_methods_supported: List[str] = []
    token_endpoint_auth_signing_alg_values_supported: List[str] = []
    display_values_supported: List[str] = []
    claim_types_supported: List[str] = []
    claims_supported: List[str] = []
    service_documentation: Optional[str] = None
    claims_locales_supported: List[str] = []
    ui_locales_supported: List[str] = []
    claims_parameter_supported: bool = False
    request_parameter_supported: bool = False
    request_uri_parameter_supported: bool = True
    require_request_uri_registration: bool = False
    op_policy_uri: Optional[str] = None
    op_tos_uri: Optional[str] = None
    end_session_endpoint: Optional[str] = None
    code_challenge_methods_supported: List[str] = []


class Scope(Enum):
    """The scope base class for determining the scope type."""

    pass


class OAuthScope(Scope):
    offlineAccess = "offline_access"


class UserInfoScope(Scope):
    """
    The available scopes for the userinfo endpoint and the ID token claims.
    """

    openid = "openid"
    """The preserved scope for OpenID Connect. It maps to the `sub` claim."""
    profile = "profile"
    """The scope for the basic profile. It maps to the `name`, `username`, `picture` claims."""
    email = "email"
    """The scope for the email address. It maps to the `email`, `email_verified` claims."""
    phone = "phone"
    """The scope for the phone number. It maps to the `phone_number`, `phone_number_verified` claims."""
    customData = "custom_data"
    """
    The scope for the custom data. It maps to the `custom_data` claim.

    Note that the custom data is not included in the ID token by default. You need to
    use `fetchUserInfo()` to get the custom data.
    """
    identities = "identities"
    """
    The scope for the identities. It maps to the `identities` claim.

    Note that the identities are not included in the ID token by default. You need to
    use `fetchUserInfo()` to get the identities.
    """


class IdTokenClaims(BaseModel):
    """
    The ID token claims object.

    To access the extra claims, use `__pydantic_extra__`. See
    https://docs.pydantic.dev/latest/usage/models/#extra-fields for more information.
    """

    model_config = ConfigDict(extra="allow")

    iss: str
    """The issuer identifier for whom issued the token."""
    sub: str
    """The subject identifier for whom the token is intended (user ID)."""
    aud: str
    """
    The audience that the token is intended for, which is the client ID.
    """
    exp: int
    """The expiration time of the token (in seconds)."""
    iat: int
    """The time at which the token was issued (in seconds)."""
    at_hash: Optional[str] = None
    name: Optional[str] = None
    """The user's full name."""
    username: Optional[str] = None
    """The user's username."""
    picture: Optional[str] = None
    """The user's profile picture URL."""
    email: Optional[str] = None
    """The user's email address."""
    email_verified: Optional[bool] = None
    """Whether the user's email address is verified."""
    phone_number: Optional[str] = None
    """The user's phone number."""
    phone_number_verified: Optional[bool] = None
    """Whether the user's phone number is verified."""


class AccessTokenClaims(BaseModel):
    """
    The access token claims object.

    To access the extra claims, use `__pydantic_extra__`. See
    https://docs.pydantic.dev/latest/usage/models/#extra-fields for more information.
    """

    model_config = ConfigDict(extra="allow")

    iss: str
    """
    The issuer identifier for whom issued the token.
    """
    sub: str
    """
    The subject identifier for whom the token is intended (user ID).
    """
    aud: str
    """
    The audience that the token is intended for, which is the client ID or the resource
    indicator.
    """
    exp: int
    """
    The expiration time of the token (in seconds).
    """
    iat: int
    """
    The time at which the token was issued (in seconds).
    """
    scope: str
    """
    The scopes that the token is granted for.
    """
    client_id: Optional[str] = None
    """
    The client ID that the token is granted for. Useful when the client ID is not
    included in the `aud` claim.
    """



"""
Logto client storage abstract class and the default implementation.
"""

from abc import ABC, abstractmethod
from typing import Dict, Literal, Optional

PersistKey = Literal["idToken", "accessTokenMap", "refreshToken", "signInSession"]
"""
The keys literal for the persistent storage.
"""


class Storage(ABC):
    """
    The storage interface for the Logto client. Logto client will use this
    interface to store and retrieve the logto session data.

    Usually this should be implemented as a persistent storage, such as a
    session or a database, since the page will be redirected to Logto and
    then back to the original page.
    """

    @abstractmethod
    def get(self, key: PersistKey) -> Optional[str]:
        """
        Get the stored string for the given key, return None if not found.
        """
        ...

    @abstractmethod
    def set(self, key: PersistKey, value: Optional[str]) -> None:
        """
        Set the stored value (string or None) for the given key.
        """
        ...

    @abstractmethod
    def delete(self, key: PersistKey) -> None:
        ...
        """
        Delete the stored value for the given key.
        """


class MemoryStorage(Storage):
    """
    The in-memory storage implementation for the Logto client. Note this should
    only be used for testing, since the data will be lost after the page is
    redirected.

    See `Storage` for the interface.
    """

    def printWarning() -> None:
        print(
            "WARNING: Using MemoryStorage for Logto client, this should only be used for testing.",
            "Replace with a persistent storage for production.",
        )

    def __init__(self) -> None:
        self._data: Dict[str, str] = {}

    def get(self, key: str) -> Optional[str]:
        MemoryStorage.printWarning()
        return self._data.get(key, None)

    def set(self, key: str, value: Optional[str]) -> None:
        MemoryStorage.printWarning()
        self._data[key] = value

    def delete(self, key: str) -> None:
        MemoryStorage.printWarning()
        self._data.pop(key, None)



from .LogtoClient import (
    LogtoClient as LogtoClient,
    LogtoConfig as LogtoConfig,
    InteractionMode as InteractionMode,
    AccessToken as AccessToken,
)
from .LogtoException import LogtoException as LogtoException
from .Storage import Storage as Storage, PersistKey as PersistKey
from .models.oidc import (
    AccessTokenClaims as AccessTokenClaims,
    IdTokenClaims as IdTokenClaims,
    OidcProviderMetadata as OidcProviderMetadata,
    Scope as Scope,
    UserInfoScope as UserInfoScope,
)
from .models.response import (
    TokenResponse as TokenResponse,
    UserInfoResponse as UserInfoResponse,
)



"""
The core OIDC functions for the Logto client. Provider-agonistic functions
are implemented as static methods, while other functions are implemented as
instance methods.
"""

import hashlib
import secrets
import aiohttp
from jwt import PyJWKClient
import jwt
from typing import List, Optional

from .LogtoException import LogtoException
from .models.oidc import (
    AccessTokenClaims,
    IdTokenClaims,
    OAuthScope,
    OidcProviderMetadata,
    Scope,
    UserInfoScope,
)
from .models.response import TokenResponse, UserInfoResponse
from .utilities import removeFalsyKeys, urlsafeEncode


class OidcCore:
    defaultScopes: List[Scope] = [
        UserInfoScope.openid,
        OAuthScope.offlineAccess,
        UserInfoScope.profile,
    ]

    def __init__(self, metadata: OidcProviderMetadata) -> None:
        """
        Initialize the OIDC core with the provider metadata. You can use the
        `getProviderMetadata` method to fetch the provider metadata from the
        discovery URL.
        """
        self.metadata = metadata
        self.jwksClient = PyJWKClient(
            metadata.jwks_uri, headers={"user-agent": "@logto/python", "accept": "*/*"}
        )

    def generateState() -> str:
        """
        Generate a random string (32 bytes) for the state parameter.
        """
        return urlsafeEncode(secrets.token_bytes(32))

    def generateCodeVerifier() -> str:
        """
        Generate a random code verifier string (32 bytes) for PKCE.

        See: https://www.rfc-editor.org/rfc/rfc7636.html#section-4.1
        """
        return urlsafeEncode(secrets.token_bytes(32))

    def generateCodeChallenge(codeVerifier: str) -> str:
        """
        Generate a code challenge string for the given code verifier string.

        See: https://www.rfc-editor.org/rfc/rfc7636.html#section-4.2
        """
        return urlsafeEncode(hashlib.sha256(codeVerifier.encode("ascii")).digest())

    def decodeIdToken(idToken: str) -> IdTokenClaims:
        """
        Decode the ID Token and return the claims without verifying the signature.
        """
        return IdTokenClaims(**jwt.decode(idToken, options={"verify_signature": False}))

    def decodeAccessToken(accessToken: str) -> AccessTokenClaims:
        """
        Decode the access token and return the claims without verifying the signature.
        """
        return AccessTokenClaims(
            **jwt.decode(accessToken, options={"verify_signature": False})
        )

    async def getProviderMetadata(discoveryUrl: str) -> OidcProviderMetadata:
        """
        Fetch the provider metadata from the discovery URL.
        """
        async with aiohttp.ClientSession() as session:
            async with session.get(discoveryUrl) as resp:
                json = await resp.json()
                return OidcProviderMetadata(**json)

    async def fetchTokenByCode(
        self,
        clientId: str,
        clientSecret: Optional[str],
        redirectUri: str,
        code: str,
        codeVerifier: str,
    ) -> TokenResponse:
        """
        Fetch the token from the token endpoint using the authorization code.
        """
        tokenEndpoint = self.metadata.token_endpoint
        async with aiohttp.ClientSession() as session:
            async with session.post(
                tokenEndpoint,
                data={
                    "grant_type": "authorization_code",
                    "client_id": clientId,
                    "client_secret": clientSecret,
                    "redirect_uri": redirectUri,
                    "code": code,
                    "code_verifier": codeVerifier,
                },
            ) as resp:
                if resp.status != 200:
                    raise LogtoException(await resp.text())

                json = await resp.json()
                return TokenResponse(**json)

    async def fetchTokenByRefreshToken(
        self,
        clientId: str,
        clientSecret: Optional[str],
        refreshToken: str,
        resource: str = "",
    ) -> TokenResponse:
        """
        Fetch the token from the token endpoint using the refresh token.
        """
        tokenEndpoint = self.metadata.token_endpoint
        async with aiohttp.ClientSession() as session:
            async with session.post(
                tokenEndpoint,
                data=removeFalsyKeys(
                    {
                        "grant_type": "refresh_token",
                        "client_id": clientId,
                        "client_secret": clientSecret,
                        "refresh_token": refreshToken,
                        "resource": resource,
                    }
                ),
            ) as resp:
                if resp.status != 200:
                    raise LogtoException(await resp.text())

                json = await resp.json()
                return TokenResponse(**json)

    def verifyIdToken(self, idToken: str, clientId: str) -> None:
        """
        Verify the ID Token signature and its issuer and client ID, throw an exception
        if the verification fails.
        """
        issuer = self.metadata.issuer
        signing_key = self.jwksClient.get_signing_key_from_jwt(idToken)
        jwt.decode(
            idToken,
            signing_key.key,
            algorithms=["RS256", "PS256", "ES256", "ES384", "ES512"],
            audience=clientId,
            issuer=issuer,
            leeway=30,
        )

    async def fetchUserInfo(self, accessToken: str) -> UserInfoResponse:
        """
        Fetch the user info from the OpenID Connect UserInfo endpoint.

        See: https://openid.net/specs/openid-connect-core-1_0.html#UserInfo
        """
        userInfoEndpoint = self.metadata.userinfo_endpoint
        async with aiohttp.ClientSession() as session:
            async with session.get(
                userInfoEndpoint, headers={"Authorization": f"Bearer {accessToken}"}
            ) as resp:
                if resp.status != 200:
                    raise LogtoException(await resp.text())

                json = await resp.json()
                return UserInfoResponse(**json)



from typing import Any, Dict, Optional
from pytest_mock import MockerFixture

from logto.models.oidc import OidcProviderMetadata


class MockResponse:
    def __init__(
        self, json: Optional[Dict[str, Any]], text: Optional[str], status: int
    ) -> None:
        self._json = json
        self._text = text or str(json)
        self.status = status

    async def json(self):
        return self._json

    async def text(self):
        return self._text

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        pass


def mockHttp(
    mocker: MockerFixture,
    method: str,
    json: Optional[Dict[str, Any]],
    text: Optional[str],
    status=200,
):
    mocker.patch(
        f"aiohttp.ClientSession.{method}",
        return_value=MockResponse(json=json, text=text, status=status),
    )


mockProviderMetadata = OidcProviderMetadata(
    issuer="https://logto.app",
    authorization_endpoint="https://logto.app/oidc/auth",
    token_endpoint="https://logto.app/oidc/auth/token",
    userinfo_endpoint="https://logto.app/oidc/userinfo",
    jwks_uri="https://logto.app/oidc/jwks",
    response_types_supported=[],
    subject_types_supported=[],
    id_token_signing_alg_values_supported=[],
)



import time
from typing import Any, Callable, Dict, Optional
from jwt import PyJWK
import jwt
from pytest_mock import MockerFixture
import pytest

from . import LogtoException
from .utilities.test import mockHttp, mockProviderMetadata
from .models.response import TokenResponse, UserInfoResponse
from .models.oidc import IdTokenClaims, AccessTokenClaims, OidcProviderMetadata
from .OidcCore import OidcCore

MockRequest = Callable[..., None]


class TestOidcCoreStatic:
    def test_generateState(self):
        assert len(OidcCore.generateState()) == 43

    def test_generateCodeVerifier(self):
        assert len(OidcCore.generateCodeVerifier()) == 43

    def test_generateCodeChallenge(self):
        codeVerifier = "12345678901234567890123456789012345678901234567890"
        assert (
            OidcCore.generateCodeChallenge(codeVerifier)
            == "9Y__uhKapn7GO_ElcaQpd8C3hdOyqTzAU4VXyR2iEV0"
        )

    def test_decodeIdToken(self):
        idToken = "eyJhbGciOiJSUzI1NiIsImtpZCI6IjEifQ.eyJpc3MiOiJodHRwczovL2xvZ3RvLmFwcCIsImF1ZCI6ImZvbyIsImV4cCI6MTYxNjQ0NjQwMCwiaWF0IjoxNjE2NDQ2MzAwLCJzdWIiOiJ1c2VyMSIsIm5hbWUiOiJKb2huIFdpY2siLCJ1c2VybmFtZSI6ImpvaG4iLCJlbWFpbCI6ImpvaG5Ad2ljay5jb20iLCJlbWFpbF92ZXJpZmllZCI6dHJ1ZX0.12345678901234567890123456789012345678901234567890"
        assert OidcCore.decodeIdToken(idToken) == IdTokenClaims(
            iss="https://logto.app",
            aud="foo",
            exp=1616446400,
            iat=1616446300,
            sub="user1",
            name="John Wick",
            username="john",
            email="john@wick.com",
            email_verified=True,
        )

    def test_decodeAccessTokenWithResourceAndClientId(self):
        accessToken = "eyJhbGciOiJSUzI1NiIsImtpZCI6IjEifQ.eyJpc3MiOiJodHRwczovL2xvZ3RvLmFwcCIsImF1ZCI6Imh0dHBzOi8vbG9ndG8uYXBwL2FwaSIsImV4cCI6MTYxNjQ0NjQwMCwiaWF0IjoxNjE2NDQ2MzAwLCJzdWIiOiJ1c2VyMSIsInNjb3BlIjoiYWRtaW4gdXNlciIsImNsaWVudF9pZCI6InNhcXJlMW9xYmtwajZ6aHE4NWhvMCJ9.12345678901234567890123456789012345678901234567890"
        assert OidcCore.decodeAccessToken(accessToken) == AccessTokenClaims(
            iss="https://logto.app",
            aud="https://logto.app/api",
            exp=1616446400,
            iat=1616446300,
            sub="user1",
            scope="admin user",
            client_id="saqre1oqbkpj6zhq85ho0",
        )


class TestOidcCore:
    @pytest.fixture
    def oidcCore(self, metadata: OidcProviderMetadata) -> OidcCore:
        return OidcCore(metadata)

    @pytest.fixture
    def metadata(self) -> OidcProviderMetadata:
        return mockProviderMetadata

    @pytest.fixture
    def tokenResponse(self) -> TokenResponse:
        return TokenResponse(access_token="token", token_type="Bearer", expires_in=3600)

    @pytest.fixture
    def mockRequest(self, mocker: MockerFixture) -> MockRequest:
        def _mock(
            method: str = "get",
            json: Optional[Dict[str, Any]] = None,
            text: Optional[str] = None,
            status: int = 200,
        ):
            return mockHttp(mocker, method, json, text, status)

        return _mock

    async def test_getProviderMetadata(
        self,
        metadata: OidcProviderMetadata,
        mockRequest: MockRequest,
    ) -> None:
        discovery_url = "https://discovery.url"
        mockRequest(json=metadata.__dict__)
        result = await OidcCore.getProviderMetadata(discovery_url)

        assert result == metadata

    async def test_fetchTokenByCode(
        self,
        oidcCore: OidcCore,
        tokenResponse: TokenResponse,
        mockRequest: MockRequest,
    ) -> None:
        mockRequest(method="post", json=tokenResponse.__dict__)
        result = await oidcCore.fetchTokenByCode(
            "clientId", "clientSecret", "redirectUri", "code", "codeVerifier"
        )

        assert result == tokenResponse

    async def test_fetchTokenByCode_failure(
        self,
        oidcCore: OidcCore,
        mockRequest: MockRequest,
    ) -> None:
        mockRequest(method="post", text="error", status=400)
        with pytest.raises(LogtoException, match="error"):
            await oidcCore.fetchTokenByCode(
                "clientId", "clientSecret", "redirectUri", "code", "codeVerifier"
            )

    async def test_fetchTokenByRefreshToken(
        self,
        oidcCore: OidcCore,
        tokenResponse: TokenResponse,
        mockRequest: MockRequest,
    ) -> None:
        mockRequest(method="post", json=tokenResponse.__dict__)
        result = await oidcCore.fetchTokenByRefreshToken(
            "clientId", "clientSecret", "refreshToken"
        )

        assert result == tokenResponse

    async def test_fetchTokenByRefreshToken_failure(
        self,
        oidcCore: OidcCore,
        mockRequest: MockRequest,
    ) -> None:
        mockRequest(method="post", text="error", status=400)
        with pytest.raises(LogtoException, match="error"):
            await oidcCore.fetchTokenByRefreshToken(
                "clientId", "clientSecret", "refreshToken"
            )

    async def test_verifyIdToken(
        self,
        oidcCore: OidcCore,
        mocker: MockerFixture,
    ) -> None:
        # Mock PyJWK with a valid key
        jwk = PyJWK(
            jwk_data={
                "kty": "EC",
                "d": "EQw2P8sukYhYuc_H8Q5pV8oTlXfAd7TM1mB4fwrYuw4BGFBcFx-Y9q5g6lvyxfG9",
                "use": "sig",
                "crv": "P-384",
                "kid": "1",
                "x": "GWEhvHiHu2nfZNn741QeWPyn3Laphn11wcD9c5LWqPQTaqw-SlJIWXavrvl4Yv7f",
                "y": "0KiYwX8U2pb74HCRby6ljlNgQGD-v_j5QN-MzXObRYa7XRQzKCrqj0_4BZN6UcS6",
                "alg": "ES384",
            }
        )

        idToken = IdTokenClaims(
            iss="https://logto.app",
            aud="foo",
            exp=int(time.time() - 10),  # Test for the leeway
            iat=1616446300,
            sub="user1",
            name="John Wick",
            username="john",
            email="john@wick.com",
            email_verified=True,
        )

        # Sign the idToken with the private key
        idTokenString = jwt.encode(
            idToken.model_dump(),
            jwk.key,
            algorithm="ES384",
            headers={"kid": "1"},
        )

        # Mock `get_signing_key_from_jwt` response from jwksClient
        mocker.patch(
            "jwt.jwks_client.PyJWKClient.get_signing_key_from_jwt",
            return_value=jwk,
        )

        # No error should be raised
        oidcCore.verifyIdToken(
            idToken=idTokenString,
            clientId="foo",
        )

    async def test_fetchUserInfo(
        self,
        oidcCore: OidcCore,
        mockRequest: MockRequest,
    ) -> None:
        mockRequest(json={"sub": "user1"})
        result = await oidcCore.fetchUserInfo("token")

        assert result == UserInfoResponse(sub="user1")

    async def test_fetchUserInfo_failure(
        self,
        oidcCore: OidcCore,
        mockRequest: MockRequest,
    ) -> None:
        mockRequest(text="error", status=400)
        with pytest.raises(LogtoException, match="error"):
            await oidcCore.fetchUserInfo("token")



from . import removeFalsyKeys, urlsafeEncode


class TestRemoveFalsyKeys:
    def test_shouldRemoveNoneAndEmptyString(self):
        assert removeFalsyKeys(
            {
                "a": None,
                "b": "",
                "c": "c",
            }
        ) == {
            "c": "c",
        }

    def test_shouldRemoveFalse(self):
        assert removeFalsyKeys(
            {
                "a": False,
                "b": True,
            }
        ) == {
            "b": True,
        }


class TestUrlsafeEncode:
    def test_shouldEncode(self):
        assert urlsafeEncode(b"123") == "MTIz"

    def test_shouldRemovePadding(self):
        assert urlsafeEncode(b"1234") == "MTIzNA"

    def test_shouldReplaceUnsafeCharacters(self):
        assert (
            urlsafeEncode(
                b"\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f"
            )
            == "AAECAwQFBgcICQoLDA0ODw"
        )

    def test_shouldReplaceUnsafeEncodedCharacters(self):
        assert urlsafeEncode(b"\xff\xff") == "__8"
